//-------------------------------------------------------------------------------------------------
// File : main.cpp
// Desc : Main Entry Point.
// Copyright(c) Project Asura. All right reserved.
//-------------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include <a3d.h>
#include <cassert>
#include <cstring>
#include <string>
#include <cstdlib>
#include <SampleApp.h>
#include <SampleUtil.h>
#include <SampleMath.h>


///////////////////////////////////////////////////////////////////////////////////////////////////
// Vertex structure
///////////////////////////////////////////////////////////////////////////////////////////////////
struct Vertex
{
    Vec3 Position;  //!< ʒuWł.
    Vec4 Color;     //!< _J[ł.
};


///////////////////////////////////////////////////////////////////////////////////////////////////
// Transform structure
///////////////////////////////////////////////////////////////////////////////////////////////////
struct Transform
{
    Mat4 World;     //!< [hsł.
    Mat4 View;      //!< r[sł.
    Mat4 Proj;      //!< ˉesł.
};


//-------------------------------------------------------------------------------------------------
// Forward declarations.
//-------------------------------------------------------------------------------------------------
bool InitA3D();
void TermA3D();
void DrawA3D();
void Resize(uint32_t w, uint32_t h, void* ptr);

//-------------------------------------------------------------------------------------------------
// Global Varaibles.
//-------------------------------------------------------------------------------------------------
IApp*                       g_pApp                  = nullptr;  //!< EBhE̐swp[NXł.
a3d::IDevice*               g_pDevice               = nullptr;  //!< foCXł.
a3d::ISwapChain*            g_pSwapChain            = nullptr;  //!< Xbv`FCł.
a3d::IQueue*                g_pGraphicsQueue        = nullptr;  //!< R}hL[ł.
a3d::IFence*                g_pFence                = nullptr;  //!< tFXł.
a3d::IDescriptorSetLayout*  g_pDescriptorSetLayout  = nullptr;  //!< fBXNv^ZbgCAEgł.
a3d::IPipelineState*        g_pPipelineState        = nullptr;  //!< pCvCXe[gł.
a3d::IBuffer*               g_pVertexBuffer         = nullptr;  //!< _obt@ł.
a3d::IBuffer*               g_pIndexBuffer          = nullptr;  //!< CfbNXobt@ł.
a3d::ITexture*              g_pDepthBuffer          = nullptr;  //!< [xobt@ł.
a3d::ITextureView*          g_pDepthView            = nullptr;  //!< [xr[ł.
a3d::ITexture*              g_pColorBuffer[2]       = {};       //!< J[obt@ł.
a3d::ITextureView*          g_pColorView[2]         = {};       //!< J[r[ł.
a3d::ICommandList*          g_pCommandList[2]       = {};       //!< R}hXgł.
a3d::IFrameBuffer*          g_pFrameBuffer[2]       = {};       //!< t[obt@ł.
a3d::IBuffer*               g_pConstantBuffer[2]    = {};       //!< 萔obt@ł.
a3d::IBufferView*           g_pConstantView[4]      = {};       //!< 萔obt@r[ł.
a3d::Viewport               g_Viewport              = {};       //!< r[|[gł.
a3d::Rect                   g_Scissor               = {};       //!< VU[`ł.
a3d::IDescriptorSet*        g_pDescriptorSet[4]     = {};       //!< fBXNv^Zbgł.
Transform                   g_Transform             = {};       //!< ϊsł.
float                       g_RotateAngle           = 0.0f;     //!< ]pł.
void*                       g_pCbHead[2]            = {};       //!< 萔obt@̐擪|C^ł.
bool                        g_Prepare               = false;    //!< łtrue.


///////////////////////////////////////////////////////////////////////////////////////////////////
// Allocator class
///////////////////////////////////////////////////////////////////////////////////////////////////
class Allocator : public a3d::IAllocator
{
    //=============================================================================================
    // list of friend classes and methods.
    //=============================================================================================
    /* NONTHING */

public:
    //=============================================================================================
    // public variables.
    //=============================================================================================
    /* NOHTING */

    //=============================================================================================
    // public methods.
    //=============================================================================================

    //---------------------------------------------------------------------------------------------
    //      mۂ܂.
    //---------------------------------------------------------------------------------------------
    void* Alloc(size_t size, size_t alignment) noexcept override
    {
        #if A3D_IS_WIN
            return _aligned_malloc(size, alignment);
        #else
            return aligned_alloc(alignment, size);
        #endif
    }

    //---------------------------------------------------------------------------------------------
    //      Ċmۂ܂.
    //---------------------------------------------------------------------------------------------
    void* Realloc(void* ptr, size_t size, size_t alignment) noexcept override
    {
        #if A3D_IS_WIN
            return _aligned_realloc(ptr, size, alignment);
        #else
            A3D_UNUSED(alignment);
            return realloc(ptr, size);
        #endif
    }

    //---------------------------------------------------------------------------------------------
    //      ܂.
    //---------------------------------------------------------------------------------------------
    void Free(void* ptr) noexcept override
    {
        #if A3D_IS_WIN
            _aligned_free(ptr);
        #else
            free(ptr);
        #endif
    }
} g_Allocator;


//-------------------------------------------------------------------------------------------------
//      CGg[|Cgł.
//-------------------------------------------------------------------------------------------------
void Main()
{
    // EBhE.
    if (!CreateApp(960, 540, &g_pApp))
    { return; }

    // TCỸR[obN֐ݒ.
    g_pApp->SetResizeCallback(Resize, nullptr);

    // A3D.
    if (!InitA3D())
    {
        g_pApp->Release();
        return;
    }

    // C[v.
    while( g_pApp->IsLoop() )
    {
        // `揈.
        DrawA3D();
    }

    // n.
    TermA3D();
    g_pApp->Release();
}

//-------------------------------------------------------------------------------------------------
//      A3D̏s܂.
//-------------------------------------------------------------------------------------------------
bool InitA3D()
{
    g_Prepare = false;

    // OtBbNXVXȅ.
    {
        a3d::SystemDesc desc = {};
        desc.pAllocator = &g_Allocator;
        desc.pOption    = g_pApp->GetWindowHandle();

        if (!a3d::InitSystem(&desc))
        { return false; }
    }

    // foCX̐.
    {
        a3d::DeviceDesc desc = {};

        desc.EnableDebug = true;

        // őfBXNv^ݒ.
        desc.MaxColorTargetCount            = 2;
        desc.MaxDepthTargetCount            = 2;
        desc.MaxShaderResourceCount         = 4;

        // őTu~bgݒ.
        desc.MaxGraphicsQueueSubmitCount    = 256;
        desc.MaxCopyQueueSubmitCount        = 256;
        desc.MaxComputeQueueSubmitCount     = 256;

        // foCX𐶐.
        if (!a3d::CreateDevice(&desc, &g_pDevice))
        { return false; }

        // R}hL[擾.
        g_pDevice->GetGraphicsQueue(&g_pGraphicsQueue);
    }

    auto info = g_pDevice->GetInfo();

    #if SAMPLE_IS_VULKAN && TARGET_PC
        auto format = a3d::RESOURCE_FORMAT_B8G8R8A8_UNORM;
    #else
        auto format = a3d::RESOURCE_FORMAT_R8G8B8A8_UNORM;
    #endif

    // Xbv`FC̐.
    {
        a3d::SwapChainDesc desc = {};
        desc.Extent.Width   = g_pApp->GetWidth();
        desc.Extent.Height  = g_pApp->GetHeight();
        desc.Format         = format;
        desc.MipLevels      = 1;
        desc.SampleCount    = 1;
        desc.BufferCount    = 2;
        desc.SyncInterval   = 1;
        desc.InstanceHandle = g_pApp->GetInstanceHandle();
        desc.WindowHandle   = g_pApp->GetWindowHandle();

        if (!g_pDevice->CreateSwapChain(&desc, &g_pSwapChain))
        { return false; }

        // Xbv`FCobt@擾.
        g_pSwapChain->GetBuffer(0, &g_pColorBuffer[0]);
        g_pSwapChain->GetBuffer(1, &g_pColorBuffer[1]);

        a3d::TextureViewDesc viewDesc = {};
        viewDesc.Dimension          = a3d::VIEW_DIMENSION_TEXTURE2D;
        viewDesc.Format             = format;
        viewDesc.TextureAspect      = a3d::TEXTURE_ASPECT_COLOR;
        viewDesc.MipSlice           = 0;
        viewDesc.MipLevels          = desc.MipLevels;
        viewDesc.FirstArraySlice    = 0;
        viewDesc.ArraySize          = 1;
        viewDesc.ComponentMapping.R = a3d::TEXTURE_SWIZZLE_R;
        viewDesc.ComponentMapping.G = a3d::TEXTURE_SWIZZLE_G;
        viewDesc.ComponentMapping.B = a3d::TEXTURE_SWIZZLE_B;
        viewDesc.ComponentMapping.A = a3d::TEXTURE_SWIZZLE_A;

        for(auto i=0; i<2; ++i)
        {
            if (!g_pDevice->CreateTextureView(g_pColorBuffer[i], &viewDesc, &g_pColorView[i]))
            { return false; }
        }
    }

    // [xobt@̐.
    {
        a3d::TextureDesc desc = {};
        desc.Dimension                      = a3d::RESOURCE_DIMENSION_TEXTURE2D;
        desc.Width                          = g_pApp->GetWidth();
        desc.Height                         = g_pApp->GetHeight();
        desc.DepthOrArraySize               = 1;
        desc.Format                         = a3d::RESOURCE_FORMAT_D32_FLOAT;
        desc.MipLevels                      = 1;
        desc.SampleCount                    = 1;
        desc.Layout                         = a3d::RESOURCE_LAYOUT_OPTIMAL;
        desc.Usage                          = a3d::RESOURCE_USAGE_DEPTH_TARGET;
        desc.InitState                      = a3d::RESOURCE_STATE_DEPTH_WRITE;
        desc.HeapProperty.Type              = a3d::HEAP_TYPE_DEFAULT;
        desc.HeapProperty.CpuPageProperty   = a3d::CPU_PAGE_PROPERTY_DEFAULT;

        if (!g_pDevice->CreateTexture(&desc, &g_pDepthBuffer))
        { return false; }

        a3d::TextureViewDesc viewDesc = {};
        viewDesc.Dimension          = a3d::VIEW_DIMENSION_TEXTURE2D;
        viewDesc.Format             = desc.Format;
        viewDesc.TextureAspect      = a3d::TEXTURE_ASPECT_DEPTH;
        viewDesc.MipSlice           = 0;
        viewDesc.MipLevels          = desc.MipLevels;
        viewDesc.FirstArraySlice    = 0;
        viewDesc.ArraySize          = desc.DepthOrArraySize;
        viewDesc.ComponentMapping.R = a3d::TEXTURE_SWIZZLE_R;
        viewDesc.ComponentMapping.G = a3d::TEXTURE_SWIZZLE_G;
        viewDesc.ComponentMapping.B = a3d::TEXTURE_SWIZZLE_B;
        viewDesc.ComponentMapping.A = a3d::TEXTURE_SWIZZLE_A;

        if (!g_pDevice->CreateTextureView(g_pDepthBuffer, &viewDesc, &g_pDepthView))
        { return false; }
    }

    // t[obt@̐
    {
        // t[obt@̐ݒ.
        a3d::FrameBufferDesc desc = {};
        desc.ColorCount         = 1;
        desc.pColorTargets[0]   = g_pColorView[0];
        desc.pDepthTarget       = g_pDepthView;

        // 1ڂ̃t[obt@𐶐.
        if (!g_pDevice->CreateFrameBuffer(&desc, &g_pFrameBuffer[0]))
        { return false; }

        // 2ڂ̃t[obt@𐶐.
        desc.pColorTargets[0] = g_pColorView[1];
        if (!g_pDevice->CreateFrameBuffer(&desc, &g_pFrameBuffer[1]))
        { return false; }
    }

    // R}hXg𐶐.
    {
        for(auto i=0; i<2; ++i)
        {
            if (!g_pDevice->CreateCommandList(a3d::COMMANDLIST_TYPE_DIRECT, &g_pCommandList[i]))
            { return false; }
        }
    }

    // tFX𐶐.
    {
        if (!g_pDevice->CreateFence(&g_pFence))
        { return false; }
    }

    // _obt@𐶐.
    {
        Vertex vertices[] = {
            { Vec3( 1.0f, -1.0f, 0.0f), Vec4(1.0f, 0.0f, 0.0f, 1.0f) },
            { Vec3(-1.0f, -1.0f, 0.0f), Vec4(0.0f, 1.0f, 0.0f, 1.0f) },
            { Vec3(-1.0f,  1.0f, 0.0f), Vec4(0.0f, 0.0f, 1.0f, 1.0f) },
            { Vec3( 1.0f,  1.0f, 0.0f), Vec4(1.0f, 0.0f, 1.0f, 1.0f) },
        };

        a3d::BufferDesc desc = {};
        desc.Size                           = sizeof(vertices);
        desc.Stride                         = sizeof(Vertex);
        desc.InitState                      = a3d::RESOURCE_STATE_GENERAL;
        desc.Usage                          = a3d::RESOURCE_USAGE_VERTEX_BUFFER;
        desc.HeapProperty.Type              = a3d::HEAP_TYPE_UPLOAD;
        desc.HeapProperty.CpuPageProperty   = a3d::CPU_PAGE_PROPERTY_DEFAULT;

        if ( !g_pDevice->CreateBuffer(&desc, &g_pVertexBuffer) )
        { return false; }

        auto ptr = g_pVertexBuffer->Map();
        if ( ptr == nullptr )
        { return false; }

        memcpy( ptr, vertices, sizeof(vertices) );
        
        g_pVertexBuffer->Unmap();
    }

    // CfbNXobt@𐶐.
    {
        uint32_t indices[] = {
            0, 1, 2,
            2, 0, 3,
        };

        a3d::BufferDesc desc = {};
        desc.Size                           = sizeof(indices);
        desc.Stride                         = sizeof(uint32_t);
        desc.InitState                      = a3d::RESOURCE_STATE_GENERAL;
        desc.Usage                          = a3d::RESOURCE_USAGE_INDEX_BUFFER;
        desc.HeapProperty.Type              = a3d::HEAP_TYPE_UPLOAD;
        desc.HeapProperty.CpuPageProperty   = a3d::CPU_PAGE_PROPERTY_DEFAULT;

        if ( !g_pDevice->CreateBuffer(&desc, &g_pIndexBuffer) )
        { return false; }

        auto ptr = g_pIndexBuffer->Map();
        if ( ptr == nullptr )
        { return false; }

        memcpy( ptr, indices, sizeof(indices) );

        g_pIndexBuffer->Unmap();
    }

    // 萔obt@𐶐.
    {
        auto stride = a3d::RoundUp<uint32_t>( sizeof(Transform), info.ConstantBufferMemoryAlignment );

        a3d::BufferDesc desc = {};
        desc.Size                           = stride * 2;
        desc.Stride                         = stride;
        desc.InitState                      = a3d::RESOURCE_STATE_GENERAL;
        desc.Usage                          = a3d::RESOURCE_USAGE_CONSTANT_BUFFER;
        desc.HeapProperty.Type              = a3d::HEAP_TYPE_UPLOAD;
        desc.HeapProperty.CpuPageProperty   = a3d::CPU_PAGE_PROPERTY_DEFAULT;

        a3d::BufferViewDesc viewDesc = {};
        viewDesc.Offset = 0;
        viewDesc.Range  = stride;

        for(auto i=0; i<2; ++i)
        {
            if (!g_pDevice->CreateBuffer(&desc, &g_pConstantBuffer[i]))
            { return false; }

            viewDesc.Offset = 0;
            if (!g_pDevice->CreateBufferView(g_pConstantBuffer[i], &viewDesc, &g_pConstantView[i * 2 + 0]))
            { return false; }

            viewDesc.Offset += stride;
            if (!g_pDevice->CreateBufferView(g_pConstantBuffer[i], &viewDesc, &g_pConstantView[i * 2 + 1]))
            { return false; }

            g_pCbHead[i] = g_pConstantBuffer[i]->Map();
        }

        Vec3 pos = Vec3(0.0f, 0.0f, -5.0f);
        Vec3 at  = Vec3(0.0f, 0.0f, 0.0f);
        Vec3 up  = Vec3(0.0f, 1.0f, 0.0f);

        auto aspect = static_cast<float>(g_pApp->GetWidth()) / static_cast<float>(g_pApp->GetHeight());

        g_Transform.World = Mat4::Identity();
        g_Transform.View  = Mat4::LookAt(pos, at, up);
        g_Transform.Proj  = Mat4::PersFov(ToRadian(45.0f), aspect, 0.1f, 1000.0f);
    }

    // VF[_oCiǂݍ݂܂.
    a3d::ShaderBinary vs = {};
    a3d::ShaderBinary ps = {};
    {
        std::string dir = GetShaderDirectoryForSampleProgram();

    #if SAMPLE_IS_VULKAN
        const auto vsFilePath = dir + "simpleVS.spv";
        const auto psFilePath = dir + "simplePS.spv";
    #elif SAMPLE_IS_D3D12 || SAMPLE_IS_D3D11
        const auto vsFilePath = dir + "simpleVS.cso";
        const auto psFilePath = dir + "simplePS.cso";
    #else
        const auto vsFilePath = dir + "simpleVS.bin";
        const auto psFilePath = dir + "simplePS.bin";
    #endif

        if (!LoadShaderBinary(vsFilePath.c_str(), vs))
        { return false; }

        if (!LoadShaderBinary(psFilePath.c_str(), ps))
        {
            DisposeShaderBinary(vs);
            return false;
        }
    }

    // fBXNv^ZbgCAEg𐶐܂.
    {
        a3d::DescriptorSetLayoutDesc desc = {};
        desc.MaxSetCount               = 4;
        desc.EntryCount                = 1;
        desc.Entries[0].ShaderMask     = a3d::SHADER_MASK_VERTEX;
        desc.Entries[0].ShaderRegister = 0;
        desc.Entries[0].Type           = a3d::DESCRIPTOR_TYPE_CBV;

        if (!g_pDevice->CreateDescriptorSetLayout(&desc, &g_pDescriptorSetLayout))
        { return false; }

        for(auto i=0; i<4; ++i)
        {
            if (!g_pDescriptorSetLayout->CreateDescriptorSet(&g_pDescriptorSet[i]))
            { return false; }

            g_pDescriptorSet[i]->SetBuffer(0, g_pConstantView[i]);
            g_pDescriptorSet[i]->Update();
        }
    }

    // OtBbNXpCvCXe[g𐶐܂.
    {
        // ͗vfł.
        a3d::InputElementDesc inputElements[] = {
            { "POSITION", 0, 0, a3d::RESOURCE_FORMAT_R32G32B32_FLOAT   , 0  },
            { "COLOR"   , 0, 1, a3d::RESOURCE_FORMAT_R32G32B32A32_FLOAT, 12 },
        };

        // ̓Xg[ł.
        a3d::InputStreamDesc inputStream = {};
        inputStream.ElementCount    = 2;
        inputStream.pElements       = inputElements;
        inputStream.StreamIndex     = 0;
        inputStream.StrideInBytes   = sizeof(Vertex);
        inputStream.InputClass      = a3d::INPUT_CLASSIFICATION_PER_VERTEX;

        // ̓CAEgł.
        a3d::InputLayoutDesc inputLayout = {};
        inputLayout.StreamCount = 1;
        inputLayout.pStreams    = &inputStream;

        // XeVeXgݒł.
        a3d::StencilTestDesc stencilTest = {};
        stencilTest.StencilFailOp      = a3d::STENCIL_OP_KEEP;
        stencilTest.StencilDepthFailOp = a3d::STENCIL_OP_KEEP;
        stencilTest.StencilFailOp      = a3d::STENCIL_OP_KEEP;
        stencilTest.StencilCompareOp   = a3d::COMPARE_OP_NEVER;

        // OtBbNXpCvCXe[gݒ肵܂.
        a3d::GraphicsPipelineStateDesc desc = {};

        // VF[_̐ݒ.
        desc.VertexShader = vs;
        desc.PixelShader  = ps;

        // uhXe[g̐ݒ.
        desc.BlendState.IndependentBlendEnable          = false;
        desc.BlendState.LogicOpEnable                   = false;
        desc.BlendState.LogicOp                         = a3d::LOGIC_OP_NOOP;
        for(auto i=0; i<8; ++i)
        {
            desc.BlendState.ColorTarget[i].BlendEnable      = false;
            desc.BlendState.ColorTarget[i].SrcBlend         = a3d::BLEND_FACTOR_ONE;
            desc.BlendState.ColorTarget[i].DstBlend         = a3d::BLEND_FACTOR_ZERO;
            desc.BlendState.ColorTarget[i].BlendOp          = a3d::BLEND_OP_ADD;
            desc.BlendState.ColorTarget[i].SrcBlendAlpha    = a3d::BLEND_FACTOR_ONE;
            desc.BlendState.ColorTarget[i].DstBlendAlpha    = a3d::BLEND_FACTOR_ZERO;
            desc.BlendState.ColorTarget[i].BlendOpAlpha     = a3d::BLEND_OP_ADD;
            desc.BlendState.ColorTarget[i].EnableWriteR     = true;
            desc.BlendState.ColorTarget[i].EnableWriteG     = true;
            desc.BlendState.ColorTarget[i].EnableWriteB     = true;
            desc.BlendState.ColorTarget[i].EnableWriteA     = true;
        }

        // X^CU\Xe[g̐ݒ.
        desc.RasterizerState.PolygonMode                = a3d::POLYGON_MODE_SOLID;
        desc.RasterizerState.CullMode                   = a3d::CULL_MODE_NONE;
        desc.RasterizerState.FrontCounterClockWise      = false;
        desc.RasterizerState.DepthBias                  = 0;
        desc.RasterizerState.DepthBiasClamp             = 0.0f;
        desc.RasterizerState.SlopeScaledDepthBias       = 0;
        desc.RasterizerState.DepthClipEnable            = false;
        desc.RasterizerState.EnableConservativeRaster   = false;
        
        // }`TvXe[g̐ݒ.
        desc.MultiSampleState.EnableAlphaToCoverage = false;
        desc.MultiSampleState.EnableMultiSample     = false;
        desc.MultiSampleState.SampleCount           = 1;

        // [xXe[g̐ݒ.
        desc.DepthState.DepthTestEnable      = true;
        desc.DepthState.DepthWriteEnable     = true;
        desc.DepthState.DepthCompareOp       = a3d::COMPARE_OP_LEQUAL;

        // XeVXe[g̐ݒ.
        desc.StencilState.StencilTestEnable    = false;
        desc.StencilState.StencllReadMask      = 0;
        desc.StencilState.StencilWriteMask     = 0;
        desc.StencilState.FrontFace            = stencilTest;
        desc.StencilState.BackFace             = stencilTest;

        // ebZ[VXe[g̐ݒ.
        desc.TessellationState.PatchControlCount = 0;

        // ̓AEg̐ݒ.
        desc.InputLayout = inputLayout;

        // fBXNv^ZbgCAEg̐ݒ.
        desc.pLayout = g_pDescriptorSetLayout;
        
        // v~eBug|W[̐ݒ.
        desc.PrimitiveTopology = a3d::PRIMITIVE_TOPOLOGY_TRIANGLELIST;

        // tH[}bg̐ݒ.
        desc.ColorCount                 = 1;
        desc.ColorTarget[0].Format      = format;
        desc.ColorTarget[0].SampleCount = 1;
        desc.DepthTarget.Format         = a3d::RESOURCE_FORMAT_D32_FLOAT;
        desc.DepthTarget.SampleCount    = 1;

        // LbVς݃pCvCXe[g̐ݒ.
        desc.pCachedPSO = nullptr;

        // OtBbNXpCvCXe[g̐.
        if (!g_pDevice->CreateGraphicsPipeline(&desc, &g_pPipelineState))
        {
            DisposeShaderBinary(vs);
            DisposeShaderBinary(ps);
            return false;
        }
    }
     // svɂȂ̂Ŕj܂.
    DisposeShaderBinary(vs);
    DisposeShaderBinary(ps);

    // r[|[g̐ݒ.
    g_Viewport.X        = 0.0f;
    g_Viewport.Y        = 0.0f;
    g_Viewport.Width    = static_cast<float>(g_pApp->GetWidth());
    g_Viewport.Height   = static_cast<float>(g_pApp->GetHeight());
    g_Viewport.MinDepth = 0.0f;
    g_Viewport.MaxDepth = 1.0f;

    // VU[`̐ݒ.
    g_Scissor.Offset.X      = 0;
    g_Scissor.Offset.Y      = 0;
    g_Scissor.Extent.Width  = g_pApp->GetWidth();
    g_Scissor.Extent.Height = g_pApp->GetHeight();

    g_Prepare = true;
    return true;
}

//-------------------------------------------------------------------------------------------------
//      A3D̏Is܂.
//-------------------------------------------------------------------------------------------------
void TermA3D()
{
    g_Prepare = false;

    // _uobt@\[X̔j.
    for(auto i=0; i<2; ++i)
    {
        // t[obt@̔j.
        a3d::SafeRelease(g_pFrameBuffer[i]);

        // J[r[̔j.
        a3d::SafeRelease(g_pColorView[i]);

        // J[obt@̔j.
        a3d::SafeRelease(g_pColorBuffer[i]);

        // R}hXg̔j.
        a3d::SafeRelease(g_pCommandList[i]);

        // 萔obt@̔j.
        a3d::SafeRelease(g_pConstantBuffer[i]);
    }

    for(auto i=0; i<4; ++i)
    {
        // 萔obt@r[̔j.
        a3d::SafeRelease(g_pConstantView[i]);

        // fBXNv^Zbg̔j.
        a3d::SafeRelease(g_pDescriptorSet[i]);
    }

    // pCvCXe[g̔j.
    a3d::SafeRelease(g_pPipelineState);

    // _obt@̔j.
    a3d::SafeRelease(g_pVertexBuffer);

    // CfbNXobt@̔j.
    a3d::SafeRelease(g_pIndexBuffer);

    // fBXNv^ZbgCAEg̔j.
    a3d::SafeRelease(g_pDescriptorSetLayout);

    // tFX̔j.
    a3d::SafeRelease(g_pFence);

    // [xXeV^[Qbgr[̔j.
    a3d::SafeRelease(g_pDepthView);

    // [xobt@̔j.
    a3d::SafeRelease(g_pDepthBuffer);

    // Xbv`FC̔j.
    a3d::SafeRelease(g_pSwapChain);

    // OtBbNXL[̔j.
    a3d::SafeRelease(g_pGraphicsQueue);

    // foCX̔j.
    a3d::SafeRelease(g_pDevice);

    // OtBbNXVXȅI.
    a3d::TermSystem();
}

//-------------------------------------------------------------------------------------------------
//      A3Dɂ`揈s܂.
//-------------------------------------------------------------------------------------------------
void DrawA3D()
{
    if (!g_Prepare)
    { return; }

    // obt@ԍ擾܂.
    auto idx = g_pSwapChain->GetCurrentBufferIndex();

    // 萔obt@XV.
    {
        g_RotateAngle += 0.025f;
        g_Transform.World = Mat4::Translation(0.0f, 0.0, 0.25f) * Mat4::RotateY(g_RotateAngle);

        auto stride = g_pConstantBuffer[idx]->GetDesc().Stride;

        auto ptr = static_cast<uint8_t*>(g_pCbHead[idx]);
        memcpy(ptr, &g_Transform, sizeof(g_Transform));

        g_Transform.World = Mat4::Scale(1.0f, 0.5f, 1.0f) * Mat4::RotateX(g_RotateAngle * 2.0f);
        ptr += stride;
        memcpy(ptr, &g_Transform, sizeof(g_Transform));
    }

    // R}h̋L^Jn܂.
    auto pCmd = g_pCommandList[idx];
    pCmd->Begin();

    // ݗp̃oAݒ肵܂.
    pCmd->TextureBarrier(
        g_pColorBuffer[idx],
        a3d::RESOURCE_STATE_PRESENT,
        a3d::RESOURCE_STATE_COLOR_WRITE);

    // t[obt@ݒ肵܂.
    pCmd->SetFrameBuffer(g_pFrameBuffer[idx]);

    // t[obt@NA܂.
    a3d::ClearColorValue clearColor = {};
    clearColor.Float[0] = 0.25f;
    clearColor.Float[1] = 0.25f;
    clearColor.Float[2] = 0.25f;
    clearColor.Float[3] = 1.0f;

    a3d::ClearDepthStencilValue clearDepth = {};
    clearDepth.Depth                = 1.0f;
    clearDepth.Stencil              = 0;
    clearDepth.EnableClearDepth     = true;
    clearDepth.EnableClearStencil   = false;

    pCmd->ClearFrameBuffer(1, &clearColor, &clearDepth);

    {
        // pCvCXe[gݒ肵܂.
        pCmd->SetPipelineState(g_pPipelineState);

        // r[|[gƃVU[`ݒ肵܂.
        // NOTE : r[|[gƃVU[`̐ݒ́CKSetPipelineState() ̌łKv܂.
        pCmd->SetViewports(1, &g_Viewport);
        pCmd->SetScissors (1, &g_Scissor);

        // _obt@ݒ肵܂.
        pCmd->SetVertexBuffers(0, 1, &g_pVertexBuffer, nullptr);
        pCmd->SetIndexBuffer(g_pIndexBuffer, 0);

        // O̎Op`.
        pCmd->SetDescriptorSet(g_pDescriptorSet[idx * 2 + 0]);
        pCmd->DrawIndexedInstanced(6, 1, 0, 0, 0);

        // ̎Op`.
        pCmd->SetDescriptorSet(g_pDescriptorSet[idx * 2 + 1]);
        pCmd->DrawIndexedInstanced(6, 1, 0, 0, 0);
    }
    
    // \p̃oAݒ肷OɁCt[obt@̐ݒKv܂.
    pCmd->SetFrameBuffer(nullptr);

    // \pɃoAݒ肵܂.
    pCmd->TextureBarrier(
        g_pColorBuffer[idx],
        a3d::RESOURCE_STATE_COLOR_WRITE,
        a3d::RESOURCE_STATE_PRESENT);

    // R}hXgւ̋L^I܂.
    pCmd->End();

    // R}hL[ɓo^܂.
    g_pGraphicsQueue->Submit(pCmd);

    // R}hs܂.
    g_pGraphicsQueue->Execute(g_pFence);

    // R}hsҋ@܂.
    if (!g_pFence->IsSignaled())
    { g_pFence->Wait(UINT32_MAX); }

    // ʂɕ\܂.
    g_pSwapChain->Present();
}

//-------------------------------------------------------------------------------------------------
//      TCYł.
//-------------------------------------------------------------------------------------------------
void Resize( uint32_t w, uint32_t h, void* pUser )
{
    A3D_UNUSED( pUser );

    // ĂȂԂCłȂ̂őI.
    if (!g_Prepare || g_pSwapChain == nullptr)
    { return; }

    // Tvȉ̏ꍇ̓NbV錴ƂȂ̂ŏȂ.
    {
        auto desc = g_pSwapChain->GetDesc();
        if ( w < desc.SampleCount || h < desc.SampleCount )
        { return; }
    }

    g_Prepare = false;

    // AChԂɂȂ܂ő҂.
    g_pGraphicsQueue->WaitIdle();
    g_pDevice->WaitIdle();

    for(auto i=0; i<2; ++i)
    {
        // t[obt@̔j.
        a3d::SafeRelease(g_pFrameBuffer[i]);

        // J[r[̔j.
        a3d::SafeRelease(g_pColorView[i]);

        // J[obt@̔j.
        a3d::SafeRelease(g_pColorBuffer[i]);
    }

    // [xXeV^[Qbgr[̔j.
    a3d::SafeRelease(g_pDepthView);

    // [xobt@̔j.
    a3d::SafeRelease(g_pDepthBuffer);

    // Xbv`FC̃TCYł.
    g_pSwapChain->ResizeBuffers( w, h );

    // eNX`r[𐶐.
    {
        auto desc = g_pSwapChain->GetDesc();

        // Xbv`FCobt@擾.
        g_pSwapChain->GetBuffer(0, &g_pColorBuffer[0]);
        g_pSwapChain->GetBuffer(1, &g_pColorBuffer[1]);

        a3d::TextureViewDesc viewDesc = {};
        viewDesc.Dimension          = a3d::VIEW_DIMENSION_TEXTURE2D;
        viewDesc.Format             = desc.Format;
        viewDesc.TextureAspect      = a3d::TEXTURE_ASPECT_COLOR;
        viewDesc.MipSlice           = 0;
        viewDesc.MipLevels          = desc.MipLevels;
        viewDesc.FirstArraySlice    = 0;
        viewDesc.ArraySize          = 1;
        viewDesc.ComponentMapping.R = a3d::TEXTURE_SWIZZLE_R;
        viewDesc.ComponentMapping.G = a3d::TEXTURE_SWIZZLE_G;
        viewDesc.ComponentMapping.B = a3d::TEXTURE_SWIZZLE_B;
        viewDesc.ComponentMapping.A = a3d::TEXTURE_SWIZZLE_A;

        for(auto i=0; i<2; ++i)
        {
            auto ret = g_pDevice->CreateTextureView(g_pColorBuffer[i], &viewDesc, &g_pColorView[i]);
            assert(ret == true);
            A3D_UNUSED(ret);
        }
    }

    // [xobt@̐.
    {
        a3d::TextureDesc desc = {};
        desc.Dimension                      = a3d::RESOURCE_DIMENSION_TEXTURE2D;
        desc.Width                          = g_pApp->GetWidth();
        desc.Height                         = g_pApp->GetHeight();
        desc.DepthOrArraySize               = 1;
        desc.Format                         = a3d::RESOURCE_FORMAT_D32_FLOAT;
        desc.MipLevels                      = 1;
        desc.SampleCount                    = 1;
        desc.Layout                         = a3d::RESOURCE_LAYOUT_OPTIMAL;
        desc.Usage                          = a3d::RESOURCE_USAGE_DEPTH_TARGET;
        desc.InitState                      = a3d::RESOURCE_STATE_DEPTH_WRITE;
        desc.HeapProperty.Type              = a3d::HEAP_TYPE_DEFAULT;
        desc.HeapProperty.CpuPageProperty   = a3d::CPU_PAGE_PROPERTY_DEFAULT;

        auto ret = g_pDevice->CreateTexture(&desc, &g_pDepthBuffer);
        assert(ret == true);
        A3D_UNUSED(ret);

        a3d::TextureViewDesc viewDesc = {};
        viewDesc.Dimension          = a3d::VIEW_DIMENSION_TEXTURE2D;
        viewDesc.Format             = desc.Format;
        viewDesc.TextureAspect      = a3d::TEXTURE_ASPECT_DEPTH;
        viewDesc.MipSlice           = 0;
        viewDesc.MipLevels          = desc.MipLevels;
        viewDesc.FirstArraySlice    = 0;
        viewDesc.ArraySize          = desc.DepthOrArraySize;
        viewDesc.ComponentMapping.R = a3d::TEXTURE_SWIZZLE_R;
        viewDesc.ComponentMapping.G = a3d::TEXTURE_SWIZZLE_G;
        viewDesc.ComponentMapping.B = a3d::TEXTURE_SWIZZLE_B;
        viewDesc.ComponentMapping.A = a3d::TEXTURE_SWIZZLE_A;

        ret = g_pDevice->CreateTextureView(g_pDepthBuffer, &viewDesc, &g_pDepthView);
        assert(ret == true);
    }

    // t[obt@̐
    {
        // t[obt@̐ݒ.
        a3d::FrameBufferDesc desc = {};
        desc.ColorCount         = 1;
        desc.pColorTargets[0]   = g_pColorView[0];
        desc.pDepthTarget       = g_pDepthView;

        // 1ڂ̃t[obt@𐶐.
        auto ret = g_pDevice->CreateFrameBuffer(&desc, &g_pFrameBuffer[0]);
        assert(ret == true);
        A3D_UNUSED(ret);

        // 2ڂ̃t[obt@𐶐.
        desc.pColorTargets[0] = g_pColorView[1];
        ret = g_pDevice->CreateFrameBuffer(&desc, &g_pFrameBuffer[1]);
        assert(ret == true);
    }

    // r[|[g̐ݒ.
    g_Viewport.X        = 0;
    g_Viewport.Y        = 0;
    g_Viewport.Width    = float(w);
    g_Viewport.Height   = float(h);

    // VU[`̐ݒ.
    g_Scissor.Offset.X      = 0;
    g_Scissor.Offset.Y      = 0;
    g_Scissor.Extent.Width  = w;
    g_Scissor.Extent.Height = h;

    g_Prepare = true;
}